/**
 * @file dots_localization.h
 * @brief Header of the localization functionality
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifndef DOTS_LOCALIZATION_H
#define DOTS_LOCALIZATION_H

#ifdef __cplusplus
extern "C" {
#endif

#include "intpoint.h"

/* * * DATA STRUCTURES * * */

///@brief Enumeration of the possible errors occuring during decoding
typedef enum {
    DECODING_NO_ERROR = 0,      //! No error occured during decoding
    DECODING_TOO_MANY_POINTS,   //! More points than expected showed up
    DECODING_NOT_ENOUGH_POINTS, //! Not enough points for a meaningful decoding
    DECODING_COLLINEAR_GRID,    //! Almost all points are aligned
    DECODING_WRONG_GRID_SIZE,   //! The detected grid-spacing is unrealistic
    DECODING_X_FAILED,          //! Decoding of the X-coordinate failed
    DECODING_Y_FAILED,          //! Decoding of the Y-coordinate failed
    DECODING_BOTH_FAILED,       //! Both decodings failed
    NO_INFO                     //! State if no decoding has taken place
} DecodingError;

/// @brief Structure which holds all the information relative to the decoding. Is used for storing the state of the decoding. Exactly one is needed by the API.
typedef struct _DotsPositionInfo{
    // state variables
    int decoded;                    //! Boolean variable indicating the success of decoding
    DecodingError decodingError;    //! Error state of the decoding
    int divisor;                    //! Fixed divisor for result

    // decoded position
    IntPoint position;              //! Decoded position*divisor
    IntPoint sector;                //! Decoded sector

    // decoded rotation
    IntPoint axes[2];               //! Grid's main directions in pixels
    RotationDecoder rotDec;         //! Rotation decoding structure holding information about the current certainties of being in the right orientation
} DotsPositionInfo;

/* * * MAIN LOCALIZATION FUNCTION * * */

///@brief  Runs the localization algorithm
///@var    previousInfo   The information of the last decoding, can be set to 0 if there is no decoding done so far
///@var    image          Must be a 8-bit grayscale image with width and height specified
///@var    rows           Number of rows (height)
///@var    cols           Number of colums (width)
///@return Pointer to the position info containing all information necessary
DotsPositionInfo* dotsLocalize(DotsPositionInfo* previousInfo, unsigned char* image, int rows, int cols);

/* * * FUNCTIONS FOR SETUP * * */

///@brief Inititializes a dotPositionInfo to default values
///@return Pointer to new DotsPositionInfo structure, NULL if allocation failed
DotsPositionInfo dotsPosInfoInit();

///@brief Gets the integer results divisor value
///@var info Pointer to an allocation DotPositionInfo structure.
///@return Divisor value
int dotsGetDivisor(DotsPositionInfo* info);

///@brief Sets the integer results divisor value
///@var info Pointer to an allocation DotPositionInfo structure.
///@var divisor New divisor, must not be 0. The divisor is essentially the scaling used on the grid coordinates.
void dotsSetDivisor(DotsPositionInfo* info, int divisor);

/* * *  POSITION INFORATION RETRIEVAL * * */

///@brief Gets X position stored in info (precision still depends on the choice of the divisor of the info-structure)
///@var info Pointer to an allocation DotPositionInfo structure.
///@return Decoded Y position of in grid coordinates as a floating point value. If no information is available: -1
float dotsGetFloatX(DotsPositionInfo* info);

///@brief Gets Y position stored in info (precision still depends on the choice of the divisor of the info-structure)
///@var info Pointer to an allocation DotPositionInfo structure.
///@return Decoded Y-position of in grid coordinates as a floating point value. If no information is available: -1
float dotsGetFloatY(DotsPositionInfo* info);

///@brief Gets X position stored in info
///@var info Pointer to an allocation DotPositionInfo structure.
///@return Decoded X position of in grid coordinates*divisor as an integer value. If no information is available: -1
int dotsGetIntX(DotsPositionInfo* info);

///@brief Gets Y position stored in info
///@var info Pointer to an allocation DotPositionInfo structure.
///@return Decoded Y position of in grid coordinates*divisor as an integer value. If no information is available: -1
int dotsGetIntY(DotsPositionInfo* info);

///@brief Gets X Sector stored in info
///@var info Pointer to an allocation DotPositionInfo structure.
///@return Decoded X Sector. If no information is available: -1
int dotsGetSectorX(DotsPositionInfo* info);

///@brief Gets Y Sector stored in info
///@var info Pointer to an allocation DotPositionInfo structure.
///@return Decoded Y Sector. If no information is available: -1
int dotsGetSectorY(DotsPositionInfo* info);

///@brief Gets the angle of the picture taking device in respect to the image
///@var info Pointer to an allocation DotPositionInfo structure.
///@return angle in radians with respect to the horizontal in counterclockwise manner [-pi,pi]
float dotsGetAngle(DotsPositionInfo* info);

// INFORMATION ABOUT DECODING STATE

///@brief Indicates if info has been decoded (DEPRECIATED)
///@var info Pointer to an allocation DotPositionInfo structure
///@return 1 if decoded 0 otherwise
int dotsIsDecoded(DotsPositionInfo *info);

///@brief Gets error code
///@var info Pointer to an allocation DotPositionInfo structure
///@return Error code of the last decoding
DecodingError dotsGetError(DotsPositionInfo* info);

///@brief Determines whether the image was taken without an underlying dot grid
///@var info Pointer to an allocation DotPositionInfo structure
///@return 1 if no grid structure is found, 0 if there was a grid structure
int dotsIsAboveGround(DotsPositionInfo *info);

#ifdef __cplusplus
}
#endif

#endif // DOTS_LOCALIZATION_H

