/**
 * @file grid.c
 * @brief Source of neighbour and edge finding
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#include "compiler.h"
#include "config.h"

#include "grid.h"

#pragma GCC push_options
#pragma GCC optimize ("O1")

#define NO_DIRECTION (-1)

/****************************************
 * Helper function forward declarations *
 ****************************************/
/// @brief  Finds the four closest neighbours for num points in pointArray
///         This method is declared in the header entirely for debugging purposes in webcam_sample
/// @var    pointArray         Vertically ordered array of point coordinates
/// @var    num                 Number of points contained in pointArray
/// @return Pointer to the neighbour Array (attention: due to the static allocation there will only be one neighborhood array)
Neighbours* findNeighbours(IntPoint* pointArray, int num);

/// @brief  Checks neighbours if they are mutual neighbours, gives back a list of mutual edges
/// @var    neigh               Pointer to array of neighborhood connections, with indices pointing to coords
/// @var    coords              Pointer to point coordinates-array
/// @var    num                 Number of entries in the neigh and coords arrays
/// @var    numEdges            Number of edges contained in the returned value
/// @return An array of edges which are mutually neighbors (attention: due to static allocation there can be only one such array)
Edge* findEdges(const Neighbours* neigh, const IntPoint* coords, int num, int* numEdges);

///@brief Takes the median of the edges and puts them into axes
///@var edges   Array of Edges with at least num entries
///@var axes    Array of 2 IntPoints, will contain the result
///@var num     number of Edges
void findInitialAxisLength(Edge* edges, IntPoint* axes, int num);

/********************************
 * Global variable declarations *
 ********************************/
static Neighbours neighbourArray[MAX_NUM_POINTS];   ///@var Holds points neighbour indices
static Edge edgeArray[MAX_NUM_POINTS*2];            ///@var Holds the links between points
static int accListPointer0[MAX_NUM_POINTS*4];       ///@var Used to calculate our corrected axes
static int accListPointer1[MAX_NUM_POINTS*4];       ///@var Used to calculate our corrected axes
static int accListPointer2[MAX_NUM_POINTS*4];       ///@var Used to calculate our corrected axes
static int accListPointer3[MAX_NUM_POINTS*4];       ///@var Used to calculate our corrected axes
static int indexGrid[MAX_NUM_POINTS];               ///@var Stores the indices associated with each grid coordinate
static int* indexRowList[MAX_NUM_POINTS];           ///@var Convenience pointer to rows of indexGrid

/*********************************
 * Core function implementations *
 *********************************/
void initAxes(IntPoint* axes, int r){
    axes[0].x = -r;
    axes[0].y = 0;
    axes[1].x = 0;
    axes[1].y = -r;
}

void findInitialGridEstimate(IntPoint* pointArray, int numPoints, IntPoint* gridAxes, IntPoint* initialOrigin){
    Neighbours* neigh = findNeighbours(pointArray, numPoints);
    int numEdges;
    Edge* edges = findEdges(neigh, pointArray, numPoints, &numEdges);
    
    //Edges will always have a positive y coordinate
    unsigned int i, it, minIdx;
    int maxDotProduct[4];
    IntPoint gridAxesExtended[4], gridAxesNormalized[4];
    gridAxesExtended[0] = gridAxes[0];
    gridAxesExtended[2] = gridAxes[0];
    intPointMul(gridAxesExtended+2, -1);
    gridAxesExtended[1] = gridAxes[1];
    gridAxesExtended[3] = gridAxes[1];
    intPointMul(gridAxesExtended+3, -1);
    IntPoint axisCentroids[4];
    unsigned int pointCounter[4];
    unsigned char winner1, winner2;

    for(it = 10; it--;){
        pointCounter[0] = axisCentroids[0].x = axisCentroids[0].y = 0;
        pointCounter[1] = axisCentroids[1].x = axisCentroids[1].y = 0;
        pointCounter[2] = axisCentroids[2].x = axisCentroids[2].y = 0;
        pointCounter[3] = axisCentroids[3].x = axisCentroids[3].y = 0;

        gridAxesNormalized[0] = intPointNormalize(gridAxesExtended[0]);
        gridAxesNormalized[1] = intPointNormalize(gridAxesExtended[1]);
        gridAxesNormalized[2] = intPointNormalize(gridAxesExtended[2]);
        gridAxesNormalized[3] = intPointNormalize(gridAxesExtended[3]);

        for(i=numEdges;i--;){
            //classify point
            int xEdge = edges[i].dx;
            int yEdge = edges[i].dy;

            maxDotProduct[0] = xEdge*gridAxesNormalized[0].x + yEdge*gridAxesNormalized[0].y;
            maxDotProduct[1] = xEdge*gridAxesNormalized[1].x + yEdge*gridAxesNormalized[1].y;
            maxDotProduct[2] = xEdge*gridAxesNormalized[2].x + yEdge*gridAxesNormalized[2].y;
            maxDotProduct[3] = xEdge*gridAxesNormalized[3].x + yEdge*gridAxesNormalized[3].y;

            winner1 = maxDotProduct[1] > maxDotProduct[0];
            winner2 = (maxDotProduct[3] > maxDotProduct[2])+2;
            if (maxDotProduct[winner1] > maxDotProduct[winner2])
                minIdx = winner1;
            else
                minIdx = winner2;

            //add classified Point to cluster
            axisCentroids[minIdx].x += xEdge;
            axisCentroids[minIdx].y += yEdge;
            ++pointCounter[minIdx];

            // save cluster info in edge
            edges[i].direction = minIdx;
        }

        //move averages one axis at a times
        int numPoint = pointCounter[0] + pointCounter[2];
        if(numPoint){
            axisCentroids[0].x = (axisCentroids[0].x - axisCentroids[2].x)/numPoint;
            axisCentroids[0].y = (axisCentroids[0].y - axisCentroids[2].y)/numPoint;
        }
        else{
            //A divide by zero axes the points are badly conditioned, or the axes are off
            //Therefore we continue the algorithm with a reinit.
            axisCentroids[0].x = -gridAxes[1].y;
            axisCentroids[0].y = gridAxes[1].x;
        }

        numPoint = pointCounter[1] + pointCounter[3];
        if(numPoint){
            axisCentroids[1].x = (axisCentroids[1].x - axisCentroids[3].x)/numPoint;
            axisCentroids[1].y = (axisCentroids[1].y - axisCentroids[3].y)/numPoint;
        }
        else{
            axisCentroids[1].x = gridAxes[0].y;
            axisCentroids[1].y = -gridAxes[0].x;
        }
        if(axisCentroids[0].x != gridAxesExtended[0].x || axisCentroids[0].y != gridAxesExtended[0].y || axisCentroids[1].x != gridAxesExtended[1].x || axisCentroids[1].y != gridAxesExtended[1].y)
            it++;
        else
            it = 0;
        gridAxesExtended[0] = axisCentroids[0];
        gridAxesExtended[1] = axisCentroids[1];
    }

    //Set length of initial axis guess to be median of edge lengths
    findInitialAxisLength(edges, gridAxesExtended, numEdges);
    gridAxes[0] = gridAxesExtended[0];
    gridAxes[1] = gridAxesExtended[1];


    //Just start origin off somewhere near the center of the image
    initialOrigin->x = pointArray[numPoints/2].x;
    initialOrigin->y = pointArray[numPoints/2].y;
}

int correctGridEstimate(IntPoint* pointArray, int numPoints, IntPoint* axes, IntPoint* origin){
    //swap vectors if order is bad
    if(axes[0].x*axes[1].y - axes[0].y*axes[1].x < 0){
        IntPoint tmp = axes[1];
        axes[1] = axes[0];
        axes[0] = tmp;
    }

    int power = SHIFT_POWER;
    int i, j, xoffset, yoffset;
    int minGridX = INT32_MAX, maxGridX = INT32_MIN, minGridY = INT32_MAX, maxGridY = INT32_MIN;
    int a = axes[0].x;// Matrix elements of matrix
    int b = axes[1].x;// a b
    int c = axes[0].y;// c d
    int d = axes[1].y;
    int determinant = a*d - c*b;
    if(!determinant)
        return 0;
    IntPoint gridCoordinate[numPoints];
    IntPoint projectedArray[numPoints];
    IntPoint* gridCoordinatePointer = gridCoordinate+numPoints;
    IntPoint* projectedArrayPtr = projectedArray+numPoints;
    IntPoint* pointPointer = pointArray+numPoints;

    //Transform to new coordinate system
    while(pointPointer != pointArray){
        --pointPointer;
        --projectedArrayPtr;
        --gridCoordinatePointer;

        //project
        xoffset = pointPointer->x - origin->x;
        yoffset = pointPointer->y - origin->y;
        projectedArrayPtr->x = (((int64_t)d*xoffset - b*yoffset) << power)/determinant;
        projectedArrayPtr->y = (((int64_t)-c*xoffset + a*yoffset) << power)/determinant;

        //assign coordinate
        gridCoordinatePointer->x = (projectedArrayPtr->x >> power) + ((projectedArrayPtr->x >> (power - 1))&1);
        if(gridCoordinatePointer->x < minGridX)
            minGridX = gridCoordinatePointer->x;
        if(gridCoordinatePointer->x > maxGridX)
            maxGridX = gridCoordinatePointer->x;
        gridCoordinatePointer->y = (projectedArrayPtr->y >> power) + ((projectedArrayPtr->y >> (power - 1))&1);
        if(gridCoordinatePointer->y < minGridY)
            minGridY = gridCoordinatePointer->y;
        if(gridCoordinatePointer->y > maxGridY)
            maxGridY = gridCoordinatePointer->y;
    }

    //Check if the points can fit in statically allocated Grid
    if(maxGridX - minGridX > MAX_GRID_LENGTH || maxGridY - minGridY > MAX_GRID_LENGTH)
        return 0;

    //Assign indices to grid coordinates to let us index into projections
    IntGrid indiceGrid = intGridCreate(maxGridX - minGridX + 1, maxGridY - minGridY + 1, indexGrid, indexRowList);
    intGridFill(&indiceGrid, NOPOINT);
    for(i=0; i<numPoints; ++i){
        indiceGrid.array[gridCoordinate[i].y - minGridY][gridCoordinate[i].x - minGridX] = i;
    }

    //Correct first axis length
    int accumulator[numPoints];
    int counter = 0;
    for(i=0; i<indiceGrid.numRows; ++i){
        for(j=0; j<indiceGrid.numCols - 1; ++j){
            int idx1 = indiceGrid.array[i][j];
            int idx2 = indiceGrid.array[i][j + 1];
            if(idx1 >= 0 && idx2 >= 0)
                accumulator[counter++] = projectedArrayPtr[idx2].x - projectedArrayPtr[idx1].x;
        }
    }
    if(counter){
        IntList tmp;
        tmp.list = accumulator;
        tmp.numElements = counter;
        int corr = intListMedian(tmp);
        axes[0].x = (a*corr) >> power;
        axes[0].y = (c*corr) >> power;
    }

    //Correct the second axis length
    counter = 0;
    for(i=0; i<indiceGrid.numRows - 1; ++i){
        for(j=0; j<indiceGrid.numCols; ++j){
            int idx1 = indiceGrid.array[i][j];
            int idx2 = indiceGrid.array[i + 1][j];
            if(idx1 >= 0 && idx2 >= 0)
                accumulator[counter++] = projectedArrayPtr[idx2].y - projectedArrayPtr[idx1].y;
        }
    }
    if(counter){
        IntList tmp;
        tmp.list = accumulator;
        tmp.numElements = counter;
        int corr = intListMedian(tmp);
        axes[1].x = (b*corr) >> power;
        axes[1].y = (d*corr) >> power;
    }

    //find a good central point for the origin
    int rowEst = indiceGrid.numRows/2;
    int colEst = indiceGrid.numCols/2;


    if(intGridFindClosestTo(indiceGrid, &rowEst, &colEst)){
        *origin = pointArray[indiceGrid.array[rowEst][colEst]];
    }

    return 1;
}

/***********************************
 * Helper function implementations *
 ***********************************/
Neighbours* findNeighbours(IntPoint* pointArray, int num){
    //This function jumps back and forth in the pointArray to find closest neighbors it stops as soon as the vertical distances make it impossible to find closer points
    unsigned int i;
    //index pointer bigger than currently pointed position
    int forward;
    //index pointer smaller than currently pointed position
    int backward;
    int dx, dy;
    unsigned int ysqrd, x, y;
    unsigned int squareDists[4], maxSquareDist, sqdist;
    Neighbours* neigb = neighbourArray;
    char winner1, winner2, maxIndex = 0;
    char looping;

    //For every index i in the pointList, we ignore the first and last for convenience/code cleanliness
    for(i=1;i<num-1;i++){
        //init distances to intmax
        squareDists[0] = UINT32_MAX;
        squareDists[1] = UINT32_MAX;
        squareDists[2] = UINT32_MAX;
        squareDists[3] = UINT32_MAX;
        maxSquareDist = UINT32_MAX;

        forward = i + 1;
        backward = i - 1;
        x = pointArray[i].x;
        y = pointArray[i].y;
        looping = 2;
        while (looping) {
            if(forward < num){
                dx = x - pointArray[forward].x;
                dy = y - pointArray[forward].y;
                ysqrd = dy*dy;
                sqdist = dx*dx + ysqrd;
                if(ysqrd >= maxSquareDist) {
                    //stop forward search, if the y-distance is bigger than the farthest neighbor's distance
                    forward = num;
                } else if (sqdist < maxSquareDist){
                    //a closer neighbor than the ones in the neigbors list is found
                    squareDists[maxIndex] = sqdist;
                    neigb[i].neigbourIdx[maxIndex] = forward;

                    //find the farthest neighbor, for the next comparisions
                    winner1 = squareDists[1] > squareDists[0];
                    winner2 = (squareDists[3] > squareDists[2])+2;
                    if (squareDists[winner2] > squareDists[winner1])
                        maxIndex = winner2;
                    else
                        maxIndex = winner1;
                    maxSquareDist = squareDists[maxIndex];
                }
                if (++forward >= num)
                    looping--;
            }
            if(backward >= 0){
                dx = x - pointArray[backward].x;
                dy = y - pointArray[backward].y;
                ysqrd = dy*dy;
                sqdist = dx*dx + ysqrd;

                if(ysqrd >= maxSquareDist) {
                    //stop backward search, if the y-distance is bigger than the farthest neighbor's distance
                    backward = -1;
                } else if(sqdist < maxSquareDist){
                    //a closer neighbor than the ones in the neigbors list is found
                    squareDists[maxIndex] = sqdist;
                    neigb[i].neigbourIdx[maxIndex] = backward;

                    //find the farthest neighbor, for the next comparisions
                    winner1 = squareDists[1] > squareDists[0];
                    winner2 = (squareDists[3] > squareDists[2])+2;
                    if (squareDists[winner2] > squareDists[winner1])
                        maxIndex = winner2;
                    else
                        maxIndex = winner1;
                    maxSquareDist = squareDists[maxIndex];
                }
                if(--backward < 0)
                    looping --;
            }
        }
    }

    return neigb;
}

Edge* findEdges(const Neighbours* neigh, const IntPoint* coords, int num, int* numEdges){
    int i, j, idx;
    int numEdge = 0;
    IntPoint curPoint, neighbour;
    Neighbours tmp;
    Edge* edges = edgeArray;
    //for each point
    for(i=num;i--;){
        curPoint = coords[i];
        //for each neighbour of point
        for(j=4;j--;){
            idx = neigh[i].neigbourIdx[j];
            if(idx > i){
                neighbour = coords[idx];
                tmp = neigh[idx];
                //for all neighbours of neighbour
                if(tmp.neigbourIdx[0] == i){
                    edges[numEdge].direction = NO_DIRECTION;
                    edges[numEdge].dx = neighbour.x - curPoint.x;
                    edges[numEdge].dy = neighbour.y - curPoint.y;
                    edges[numEdge].start = i;
                    edges[numEdge++].end = idx;
                } else if(tmp.neigbourIdx[1] == i){
                    edges[numEdge].direction = NO_DIRECTION;
                    edges[numEdge].dx = neighbour.x - curPoint.x;
                    edges[numEdge].dy = neighbour.y - curPoint.y;
                    edges[numEdge].start = i;
                    edges[numEdge++].end = idx;
                } else if(tmp.neigbourIdx[2] == i){
                    edges[numEdge].direction = NO_DIRECTION;
                    edges[numEdge].dx = neighbour.x - curPoint.x;
                    edges[numEdge].dy = neighbour.y - curPoint.y;
                    edges[numEdge].start = i;
                    edges[numEdge++].end = idx;
                } else if(tmp.neigbourIdx[3] == i){
                    edges[numEdge].direction = NO_DIRECTION;
                    edges[numEdge].dx = neighbour.x - curPoint.x;
                    edges[numEdge].dy = neighbour.y - curPoint.y;
                    edges[numEdge].start = i;
                    edges[numEdge++].end = idx;
                }
            }
        }
    }
    *numEdges = numEdge;
    return edges;
}

void findInitialAxisLength(Edge* edges, IntPoint* axes, int num){
    int dx, dy, dir;
    Edge* e;
    int squareLength[4];
    IntList accumulators[4];

    squareLength[0] = axes[0].x*axes[0].x + axes[0].y*axes[0].y;
    accumulators[0] = intListCreate(accListPointer0);
    squareLength[1] = axes[1].x*axes[1].x + axes[1].y*axes[1].y;
    accumulators[1] = intListCreate(accListPointer1);
    squareLength[2] = axes[2].x*axes[2].x + axes[2].y*axes[2].y;
    accumulators[2] = intListCreate(accListPointer2);
    squareLength[3] = axes[3].x*axes[3].x + axes[3].y*axes[3].y;
    accumulators[3] = intListCreate(accListPointer3);

    //sort edges into accumulators
    for(e=edges;e<edges+num;++e){
        dir = e->direction;
        if(dir < 2){
            dx = e->dx-axes[dir].x;
            dy = e->dy-axes[dir].y;
        }else{
            dx = e->dx + axes[dir - 2].x;
            dy = e->dy + axes[dir - 2].y;
        }

        //only consider edges a subset of edges that are relatively similar to the initial axes
        if((dx*dx + dy*dy) < 0.25*squareLength[dir]){
            intListPush(accumulators + (dir&1), (1 - (dir&2))*e->dx);
            intListPush(accumulators + (dir&1) + 2, (1 - (dir&2))*e->dy);
        }
    }

    //Each axis lengths is adjusted to be the median of the edge lengths of edges corresponding to that axis direction
    axes[0].x = intListMedian(accumulators[0]);
    axes[0].y = intListMedian(accumulators[2]);
    axes[1].x = intListMedian(accumulators[1]);
    axes[1].y = intListMedian(accumulators[3]);
}

#pragma GCC pop_options
