/**
 * @file grid.h
 * @brief Header of neighbour and edge finding
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifndef GRID_H
#define GRID_H

#ifdef __cplusplus
extern "C" {
#endif

#include "intgrid.h"
#include "intlist.h"
#include "intpoint.h"

/// @brief Structure with indices of the four closest neighbors
typedef struct _Neighbours{
    short neigbourIdx[4]; //! Array of indices to the four closest neighbors
} Neighbours;

/// @brief Structure holding all necessary information for edges
typedef struct _Edge{
    unsigned short start;      //! Index of higher point
    unsigned short end;        //! Index of lower point
    int dx;         //! Length of the edge projected to the x-axis
    int dy;         //! Length of the edge projected to the y-axis
    char direction;  //! Indication of direction
} Edge;

///@brief   Initializes axes to be 90° offset to each other and of length r
///@var     axes               Pointer to two-element array representing the grid-directions
///@var     r                  Length to be given to the grid-directions
void initAxes(IntPoint* axes, int r);

///@brief  Finds an initial estimate for the axes
///@var    pointArray              Pointer to array of IntPoint coordinates, in the same order as neigh
///@var    numPoints               Length of neigh and coords arrays
///@var    gridAxes                Pointer to array of 2 IntPoints containing an initial guess of the directions
///@var    initialOrigin           Pointer to an IntPoint, receives coordinates of a point close to the center of the image
void findInitialGridEstimate(IntPoint* pointArray, int numPoints, IntPoint* gridAxes, IntPoint* initialOrigin);

/// @brief Given estimated axes this function corrects their length.
/// It also provides the point of origin close to the center of points and aligned with the grid.
///
/// @var pointArray     Pointer to a list of the underlying PointCoordinates
/// @var numPoints      number of points in pointArray
/// @var axes           Pointer to an array of two IntPoints. Will hold the initial guess of the grid main-vectors.
/// @var origin         Will hold an IntPoint which will be the initial guess of the origin
/// @return             Success of the operation
int correctGridEstimate(IntPoint* pointArray, int numPoints, IntPoint* axes, IntPoint* origin);

/// This method is declared in the header entirely for debugging purposes in webcam_sample
Neighbours* findNeighbours(IntPoint* pointArray, int num);


#ifdef __cplusplus
}
#endif

#endif // GRID_H

