/**
 * @file gridprobability.c
 * @brief Source responsible for converting a point list to an ordered grid
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifdef __cplusplus
extern "C" {
#endif

#include "compiler.h"
#include "config.h"

#include "gridprobability.h"

#pragma GCC push_options
#pragma GCC optimize ("O1")

/********************************
 * Global variable declarations *
 ********************************/
static int xOffsetList[MAX_NUM_POINTS];             ///@var Stores our point's X-offsets with respect to the grid for use in calculculating probGrid1
static int yOffsetList[MAX_NUM_POINTS];             ///@var Stores our point's X-offsets with respect to the grid for use in calculculating probGrid2
static IntPoint gridCoordinateList[MAX_NUM_POINTS]; ///@var Grid coordinate indices for each point
static int listCopy[MAX_NUM_POINTS*4];              ///@var Allocated space use in correcting the origin location

static int probGridX[MAX_NUM_POINTS];               ///@var Probablities in one axis direction (usually X)
static int probGridY[MAX_NUM_POINTS];               ///@var Probablities in the other axis direction (usually Y)
static int probGridMax[MAX_NUM_POINTS];             ///@var Maximum of probabilities

static int* rowList1[MAX_GRID_LENGTH];              ///@var Convenience pointer to rows of probGrid1
static int* rowList2[MAX_GRID_LENGTH];              ///@var Convenience pointer to rows of probGrid2
static int* rowListMax[MAX_GRID_LENGTH];            ///@var Convenience pointer to rows of probGridMax

static IntPoint probabilityArray[MAX_NUM_POINTS];   ///@var Temporary storage for probability grids

/****************************************
 * Helper function forward declarations *
 ****************************************/
///@brief Creates an empty GridOffsetsInfo
///@return A new empty GridOffsetsInfo
GridOffsetsInfo gridOffsetsInfoInitialize();

///@brief   Calculates the median displacement between dots and their corresponding grid coordinates. Corrects the offset information.
///@var     g Completely established GridOffsetsInfo
///@return  the median (x and y) of the offsets
IntPoint gridOffsetsMedian(GridOffsetsInfo *g);

///@brief  Constructs an array of probabilities out of arrays of offsets
///@var    offsetsx    pointer to list of offsets in gridcoordinates
///@var    offsetsy    pointer to list of offsets in gridcoordinates
///@var    n           number of offsets contained in the arrays
///@return pointer to an array of length 1 containing the x, and the y- probabilities
IntPoint* probabilities(int* offsetsx, int* offsetsy, int n);

///@brief  Determines a probability value for a given offset
///@var    diagonalOffset   the cummulative offsets to the coordinate axis (u,v) (can be seen as L1-norm)
///@return A value proportional to the probabilities of the symbol being of value 1
int offsetProbability(int diagonalOffset);

/*********************************
 * Core function implementations *
 *********************************/
GridOffsetsInfo calculateGridOffsets(IntPoint* pointArray, int numPoints, IntPoint* axes, IntPoint* origin){
    GridOffsetsInfo gridOffsetsInfo = gridOffsetsInfoInitialize();
    int i, xoffset, yoffset, xproj, yproj, minGridX = INT32_MAX, maxGridX = INT32_MIN, minGridY = INT32_MAX, maxGridY = INT32_MIN;
    int power = SHIFT_POWER;
    IntPoint medianOffset;
    //matrix [a b ;c d];
    int a = axes[0].x;
    int b = axes[1].x;
    int c = axes[0].y;
    int d = axes[1].y;
    int determinant = a*d - c*b;
    if(!determinant || numPoints < 1)
        return gridOffsetsInfo;

    gridOffsetsInfo.numElements = numPoints;

    IntPoint* pointOrig = pointArray;
    int* xoffsets = gridOffsetsInfo.xoffsets;
    int* yoffsets = gridOffsetsInfo.yoffsets;
    IntPoint* gridCoordinate = gridOffsetsInfo.gridCoordinate;
    while(pointOrig <= &pointArray[numPoints - 1]){
        //project each point onto the grid
        xoffset = pointOrig->x - origin->x;
        yoffset = pointOrig->y - origin->y;
        xproj = ((int64_t)(d*xoffset - b*yoffset) << power)/determinant;
        yproj = ((int64_t)(-c*xoffset + a*yoffset) << power)/determinant;

        //assign a grid coordinate to each point
        gridCoordinate->x = (xproj >> power) + ((xproj >> (power - 1))&1);
        gridCoordinate->y = (yproj >> power) + ((yproj >> (power - 1))&1);

        //update minMax
        if(gridCoordinate->x < minGridX)
            minGridX = gridCoordinate->x;
        if(gridCoordinate->x > maxGridX)
            maxGridX = gridCoordinate->x;
        if(gridCoordinate->y < minGridY)
            minGridY = gridCoordinate->y;
        if(gridCoordinate->y > maxGridY)
            maxGridY = gridCoordinate->y;

        //find offsets:
        *xoffsets = xproj - (gridCoordinate->x << power);
        *yoffsets = yproj - (gridCoordinate->y << power);

        //Increment pointers;
        ++pointOrig;
        ++xoffsets;
        ++yoffsets;
        ++gridCoordinate;
    }
    gridOffsetsInfo.gridMaxRows = maxGridX - minGridX + 1;
    gridOffsetsInfo.gridMaxCols = maxGridY - minGridY + 1;

    //Shift coordinates to be positive.
    for(gridCoordinate = gridOffsetsInfo.gridCoordinate; gridCoordinate < gridOffsetsInfo.gridCoordinate + gridOffsetsInfo.numElements; ++gridCoordinate){
        intPointAdd(gridCoordinate, -minGridX, -minGridY);
    }

   //Median of offsets should be 0, adjust our offsets, points and origin by the actual median offset to make it so
   medianOffset = gridOffsetsMedian(&gridOffsetsInfo);
   for(i=0; i<gridOffsetsInfo.numElements; ++i){
       gridOffsetsInfo.xoffsets[i] -= medianOffset.x;
       gridOffsetsInfo.yoffsets[i] -= medianOffset.y;
       pointOrig--;
       pointOrig->x += (a*medianOffset.x + b*medianOffset.y) >> power;
       pointOrig->y += (c*medianOffset.x + d*medianOffset.y) >> power;
   }

    gridOffsetsInfo.gridOrigin = *origin;
    intPointAdd(&(gridOffsetsInfo.gridOrigin), a*minGridX + b*minGridY, c*minGridX + d*minGridY);
    intPointAdd(&(gridOffsetsInfo.gridOrigin), (a*medianOffset.x + b*medianOffset.y) >> power, (c*medianOffset.x + d*medianOffset.y) >> power);

    return gridOffsetsInfo;
}

ProbabilityGrids makeProbabilityGrids(GridOffsetsInfo gridOffsetsInfo){
    ProbabilityGrids probGrids;
    IntPoint* probs = probabilities(gridOffsetsInfo.xoffsets, gridOffsetsInfo.yoffsets, gridOffsetsInfo.numElements);
    probGrids.probX = intGridCreate(gridOffsetsInfo.gridMaxCols, gridOffsetsInfo.gridMaxRows, probGridX, rowList1);
    intGridFill(&probGrids.probX, 0);
    probGrids.probY = intGridCreate(gridOffsetsInfo.gridMaxCols, gridOffsetsInfo.gridMaxRows, probGridY, rowList2);
    intGridFill(&probGrids.probY, 0);
    probGrids.maxProb = intGridCreate(gridOffsetsInfo.gridMaxCols, gridOffsetsInfo.gridMaxRows, probGridMax, rowListMax);
    intGridFill(&probGrids.maxProb, 0);

    int probProduct, i;
    IntPoint coord;
    for(i=0; i<gridOffsetsInfo.numElements; ++i){
        probProduct = abs(probs[i].x*probs[i].y);
        coord = gridOffsetsInfo.gridCoordinate[i];
        if(probProduct >= probGrids.maxProb.array[coord.x][coord.y]){
            probGrids.maxProb.array[coord.x][coord.y] = probProduct;
            probGrids.probX.array[coord.x][coord.y] = probs[i].x;
            probGrids.probY.array[coord.x][coord.y] = probs[i].y;
        }
    }

    return probGrids;
}

/***********************************
 * Helper function implementations *
 ***********************************/
GridOffsetsInfo gridOffsetsInfoInitialize(){
    GridOffsetsInfo gridOffsetsInfo;

    gridOffsetsInfo.numElements = 0;
    gridOffsetsInfo.gridMaxCols = 0;
    gridOffsetsInfo.gridMaxRows = 0;
    gridOffsetsInfo.xoffsets = xOffsetList;
    gridOffsetsInfo.yoffsets = yOffsetList;
    gridOffsetsInfo.gridCoordinate = gridCoordinateList;

    return gridOffsetsInfo;
}

IntPoint gridOffsetsMedian(GridOffsetsInfo* g){
    IntPoint median;
    IntList tmp, cpy;
    int i;

    tmp.numElements = g->numElements;

    tmp.list = g->xoffsets;

    //used in order to not destroy the order of the list
    cpy = intListCopy(&tmp, listCopy);
    median.x = intListMedian(cpy);

    tmp.list = g->yoffsets;
    cpy = intListCopy(&tmp, listCopy);
    median.y = intListMedian(cpy);

    return median;
}

IntPoint* probabilities(int* offsetsx, int* offsetsy, int n){
    int i;
    for(i=0; i<n; ++i){
        probabilityArray[i].x = offsetProbability(*offsetsx + *offsetsy);
        probabilityArray[i].y = offsetProbability(*offsetsx - *offsetsy);
        ++offsetsx;
        ++offsetsy;
    }
    return probabilityArray;
}

#define POWERSHIFT 21 /* SHIFT USED TO AVOID FLOATS */
#define FACTOR_B 3L*MEAN_OFFSET*MEAN_OFFSET
#define FACTOR_A (-(3*PROB_MAX<<POWERSHIFT)/(2*MEAN_OFFSET*FACTOR_B))
#define HORIZONTAL_MAX_SQUARED FACTOR_B

int offsetProbability(int diagonalOffset){
    int prob = 0;

    int diagonalOffset2 = diagonalOffset*diagonalOffset;
    int factorAOffset = diagonalOffset*FACTOR_A;
    if(diagonalOffset2 < HORIZONTAL_MAX_SQUARED){
        prob = (factorAOffset*diagonalOffset2 - factorAOffset*FACTOR_B) >> POWERSHIFT;
    }

    return prob;
}

#ifdef __cplusplus
}
#endif

#pragma GCC pop_options
