/**
 * @file gridprobability.h
 * @brief Header responsible for the transmission between a point list and an ordered grid
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifndef GRID_PROBABILITY_H
#define GRID_PROBABILITY_H

#ifdef __cplusplus
extern "C" {
#endif

#include "intgrid.h"
#include "intlist.h"
#include "intpoint.h"

///@brief Stores information about our new grid coordinate system and the offsets of each
/// point relative to their grid coordinate location.
typedef struct _GridOffsetsInfo{
    int numElements;            //! Number of dots
    int gridMaxRows;            //! Number of rows
    int gridMaxCols;            //! Number of cols
    IntPoint gridOrigin;        //! Pixel-coordinates of the origin of the grid
    int* xoffsets;              //! Pointer to an array of x-offset positions
    int* yoffsets;              //! Pointer to an array of y-offset positions
    IntPoint* gridCoordinate;   //! Pointer to an array of grid coordinates
} GridOffsetsInfo;

///@brief Holds all probability grids needed for decoding
typedef struct _ProbabilityGrids{
    IntGrid maxProb; //! The absolute value of the product between probX and probY (tells the quality of the point)
    IntGrid probX;   //! Probabilities of value 1 (usually represents X, but if the grid is read the wrong way, it can be others)
    IntGrid probY;   //! Probabilities of value 2 (usually represents Y, but if the grid is read the wrong way, it can be others)
} ProbabilityGrids;

/// @brief  Initializes the new grid coordinate system, including locations of each grid coordinate an
/// @var    pointArray  Pointer to an array of the underlying PointCoordinates
/// @var    numPoints   Length of the pointArray
/// @var    axes        Pointer to an array of two InPoints. Is the initial guess of the grid main-vectors.
/// @var    origin      Initial guess of the origin in pixel coordinates
GridOffsetsInfo calculateGridOffsets(IntPoint* pointArray, int numPoints, IntPoint* axes, IntPoint* origin);

/// @brief Turns the probability grids according to the number of turns and sets the probablities according to reading direction
/// @var    grids           Pointer to a ProbabilityGrids structure (value might change)
/// @var    numRightTurns   Number of right turns to be made to the grids structure.
void probabilityGridsTurn(ProbabilityGrids* grids, int numRightTurns);

///@brief   Constructs probability grids from the dotInfo
///@var     gridOffsetsInfo     Structure holding all informations related to the dots coordinates and offsets
///@return  A ProbabilityGrids structure filled with all information
ProbabilityGrids makeProbabilityGrids(GridOffsetsInfo gridOffsetsInfo);

#ifdef __cplusplus
}
#endif

#endif // GRIDPROBABILITY_H

