/**
 * @file imgproc.c
 * @brief Source of the process to extract the points to be used for localization
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#include "compiler.h"
#include "config.h"

#include "imgproc.h"

#pragma GCC push_options
#pragma GCC optimize ("O1")

#define NOSEGMENT MAX_NUM_IMAGE_SEGMENTS
#define NOBLOB INT32_MAX
#define VISITED NOBLOB-1

/********************************
 * Global variable declarations *
 ********************************/

///@var Static array containing the center of the detected dots in (pixel coordinates in fixed comma representation)
static IntPoint pointArray[MAX_NUM_POINTS];

///@var Static array containing the detected image segments
static ImgSegment segmentList[MAX_NUM_IMAGE_SEGMENTS];

///@var Static array containg the indices of equivalence. if the value is always smaller or equal to the index.
static unsigned short equivalenceList[MAX_NUM_IMAGE_SEGMENTS];

/*********************************
 * Core function implementations *
 *********************************/
unsigned char imageMean(unsigned char* image, int rows, int cols){
    unsigned int i;
    unsigned int buffer = 0;
    unsigned int numpixel = rows*cols;
    unsigned char *imgPointer = image;
    // Assumes that numbers of pixels in each image are always multiples of 10
    for (i = numpixel; i-=10;) {
        buffer +=
        imgPointer[0] +
        imgPointer[1] +
        imgPointer[2] +
        imgPointer[3] +
        imgPointer[4] +
        imgPointer[5] +
        imgPointer[6] +
        imgPointer[7] +
        imgPointer[8] +
        imgPointer[9];
        imgPointer += 10;
    }

    return (unsigned char)(buffer/numpixel);
}

ImgSegmentList detectPoints(unsigned char* imgPointer, unsigned short nRows, unsigned short nCols, unsigned char thresholdValue){
    unsigned short i, j;
    unsigned short numSegments = 0;
    unsigned char* p;

    //init oldline
    unsigned short line1[nCols];
    unsigned short line2[nCols];

    unsigned short* oldline = line1;
    unsigned short* newline = line2;
    unsigned short* tmpline = 0;

    unsigned short last, up;

    ImgSegment foundSegments[MAX_NUM_IMAGE_SEGMENTS];

    // This code runs a basic blob detection where we loop from left to right, up to down across the image
    // and look for connected groups pixels that meet are below our color brightness threshold.
    for (j=nCols; j--;)
        oldline[j] = NOSEGMENT;

    for(i = 0; i<nRows && numSegments<MAX_NUM_IMAGE_SEGMENTS; ++i){
        p = &imgPointer[i*nCols]; //point to next row

        //Special case for first pixel of row
        if(p[0] < thresholdValue){
            up = oldline[0];
            if(up != NOSEGMENT){
                newline[0] = up;
                imgSegmentAddPixel(&foundSegments[up], 0, i);
            }
            else{
                foundSegments[numSegments] = imgSegmentCreate(0,i,1);
                newline[0] = numSegments;
                equivalenceList[numSegments] = numSegments;
                numSegments++;
                if(numSegments == MAX_NUM_IMAGE_SEGMENTS)
                    break;
            }
        }
        else
            newline[0]=NOSEGMENT;

        //case for the rest of the pixels
        for (j=1; j<nCols; ++j){
            if(p[j] < thresholdValue){
                last = newline[j-1];
                up = oldline[j];
                if(up != NOSEGMENT){
                    newline[j] = up;
                    //add to existing segment
                    imgSegmentAddPixel(&foundSegments[up], j, i);
                    //merge segment to the left with the current one
                    if(last != NOSEGMENT && last != up)
                        equivalenceList[last] = up;
                }
                else if(last != NOSEGMENT){
                    //add to existing segment
                    newline[j] = last;
                    imgSegmentAddPixel(&foundSegments[last], j, i);
                }
                else{
                    //no neighbouring segment pixels foubd, start a new segment
                    foundSegments[numSegments] = imgSegmentCreate(j, i, 1);
                    newline[j] = numSegments;
                    equivalenceList[numSegments] = numSegments;
                    numSegments++;
                    if(numSegments >= MAX_NUM_IMAGE_SEGMENTS)
                        break;
                }
            }
            else
                newline[j] = NOSEGMENT;
        }
        tmpline = oldline;
        oldline = newline;
        newline = tmpline;
        tmpline = 0;

    }

    //Equivalence Solving
    //Combine segments that have the same equivalence list value, the original segment takes precedence
    for(i = numSegments; i--;){
        if (equivalenceList[i] != i){
           imgSegmentAdd(&foundSegments[equivalenceList[i]], foundSegments[i]);
       }
    }

    //Now go through the list and pick out the original segments that are larger than the minimum size
    //This is faster than having to maintain an order of precedence in the blob detection inner loop
    unsigned short actualNumSegments = 0;
    for(i = 0; i<numSegments;i++) {
        ImgSegment foundSegment = foundSegments[i];
        if (equivalenceList[i] == i && foundSegment.n >= MIN_PIXELS_PER_SEGMENT)
            segmentList[actualNumSegments++] = foundSegment;
    }

    ImgSegmentList segList;
    segList.list = segmentList;
    segList.numElements = actualNumSegments;
    return segList;
}

IntPoint* sortPoints(ImgSegmentList list){
    int i, j;
    int num = list.numElements;
    int subdivision = SUBDIVISION;
    if(list.numElements > MAX_NUM_POINTS)
        return 0;

    //Sort list using insertion sort (~O(n))
    for(i=0; i<num; ++i){
        j = i;
        IntPoint p;
        p.x = (list.list[i].x*subdivision)/list.list[i].n;
        p.y = (list.list[j].y*subdivision)/list.list[i].n;
        pointArray[i] =  p;
        while(j > 0 && pointArray[j].y < pointArray[j - 1].y){
            IntPoint temp = pointArray[j];
            pointArray[j] = pointArray[j - 1];
            pointArray[j - 1] = temp;
            --j;
        }
    }
    return pointArray;
}

void undistortPoints(IntPoint* pointList, int numPoint, unsigned short nRows, unsigned short nCols, float k){
    if(k == 0)
        return;

    int i;
    int rSquared;
    int subdivision = SUBDIVISION;
    int largestRSquared = (nCols*nCols/4 + nRows*nRows/4)/subdivision;

    //Calculate 1/K in r^2 scale
    int realK = (int)(largestRSquared/k);

    IntPoint center = {nCols/2*subdivision, nRows/2*subdivision};
    IntPoint diff;
    for(i=0; i<numPoint; ++i){
        diff = intPointDiff(pointList[i], center);

        rSquared = (diff.x*diff.x + diff.y*diff.y)/subdivision/subdivision/subdivision;

        pointList[i].x = diff.x*realK/(rSquared + realK) + center.x;
        pointList[i].y = diff.y*realK/(rSquared + realK) + center.y;
    }
}

#pragma GCC pop_options
