/**
 * @file imgproc.h
 * @brief Header of the process to extract the maxima of an image
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifndef IMGPROC_H
#define IMGPROC_H

#include "imgsegment.h"
#include "intpoint.h"

#ifdef __cplusplus
extern "C" {
#endif

///@brief Structure holding a list of image segments
typedef struct {
    unsigned short numElements; //! Number of elements in the list
    ImgSegment* list;           //! Pointer to start of list
} ImgSegmentList;

/// @brief Calculates the mean of the image
/// @var    image           Pointer to an 8-bit image
/// @var    rows            Number of rows
/// @var    cols            Number of columns
/// @return Mean grayscale pixel value of the image
unsigned char imageMean(unsigned char* image, int rows, int cols);

/// @brief Finds 4-connected components of the image, returns at most MAX_NUM_SEGMENTS ImgSegments
/// @var    imgPointer      Pointer to an 8-bit image
/// @var    nRows           Number of rows of pixels of the image
/// @var    nCols           Number of columns of pixels of the image
/// @var    thresholdValue  Deciding value for foreground or background
/// @return An ImgSegmentList structure containing the center of all 4-connected foreground elements where each element represents a point
ImgSegmentList detectPoints(unsigned char* imgPointer, unsigned short nRows, unsigned short nCols, unsigned char thresholdValue);

/// @brief Sorts the image list according to the y-values
/// @var    list            Structure holding the information about the centers and width of the dots
/// @return An array of horizontally orederd intpoints, the list is of length list.numsegments
IntPoint* sortPoints(ImgSegmentList list);

/// @brief  Transforms every point in the pointlist according to its distance to the center
/// @var    pointList       Pointer to an array of Intpoints (array is modified)
/// @var    numPoints       Number of entries in pointList
/// @var    nRows           Number of rows in image
/// @var    nCols           Number of columns in image
/// @var    k               Distortion factor of the second degree distortion
void undistortPoints(IntPoint* pointList, int numPoint, unsigned short nRows, unsigned short nCols, float k);

#ifdef __cplusplus
}
#endif

#endif // IMGPROC_H
