/**
 * @file intgrid.c
 * @brief Contains a structure representing a matrix of integers
 * @author Lukas Hostettler
 * @author Roger Fong
 * @version 1.1
 * @date 2015-12-15
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <limits.h>
#include "compiler.h"
#include "config.h"

#include "intgrid.h"

#pragma GCC push_options
#pragma GCC optimize ("O1")

IntGrid intGridCreateEmpty(){
    IntGrid emptyGrid = {0, 0, 0, 0};
    return emptyGrid;
}

IntGrid intGridCreate(int numCol, int numRow, int* buffer, int** rowList){
    unsigned int i;
    IntGrid grid = intGridCreateEmpty();
    if(numCol && numRow){
        grid.data = buffer;
        grid.array = rowList;
        for(i=numRow; i--;){
            grid.array[i] = &grid.data[i*numCol];
        }
        if(grid.data && grid.array){
            grid.numCols = numCol;
            grid.numRows = numRow;
        }
    }
    return grid;
}

void intGridFill(IntGrid* g, int value){
    int i;
    int numPoints = g->numCols*g->numRows;
    for(i=numPoints; i--;){
        g->data[i] = value;
    }
}

void intGridCopy(IntGrid* destination, IntGrid* source) {
    memcpy(destination->data, source->data, sizeof(int)*MAX_NUM_POINTS);
    memcpy(destination->array, source->array, sizeof(int*)*MAX_GRID_LENGTH);
}

int intGridMaximum(IntGrid* g, int* idx){
    int max = INT_MIN;
    int maxIdx = 0, i;
    int numGridPoints = g->numCols*g->numRows;
    for(i=numGridPoints; i--;){
        if(g->data[i] > max){
            max = g->data[i];
            maxIdx = i;
        }
    }
    if(idx)
        *idx = maxIdx;
    return max;
}

void intGridIndToSub(IntGrid *g, int ind, int* col, int* row){
    int numCol = g->numCols;
    int numRow = g->numRows;
    if(numCol && numRow){
        *row = ind/numCol;
        *col = ind - (*row)*numCol;
    }
}

IntGrid intGridTurn(IntGrid grid, int numRightTurns, int factor, int* gridList, int** rowList){
    int i,j;
    numRightTurns %= 4;
    IntGrid tmpGrid = intGridCreateEmpty();

    switch(numRightTurns){
    case 0:
        break;
    case 1:
        tmpGrid = intGridCreate(grid.numRows, grid.numCols, gridList, rowList);
        for(i=0; i<grid.numCols; ++i){
            for(j=0; j<grid.numRows; ++j){
                tmpGrid.array[i][j] = factor*grid.array[grid.numRows - 1 - j][i];
            }
        }
        break;
    case 2:
        tmpGrid = intGridCreate(grid.numCols, grid.numRows, gridList, rowList);
        for(i=0; i<grid.numRows; ++i){
            for(j=0; j<grid.numCols; ++j){
                tmpGrid.array[i][j] = factor*grid.array[grid.numRows - 1 - i][grid.numCols - 1 - j];
            }
        }
        break;
    case 3:
        tmpGrid = intGridCreate(grid.numRows, grid.numCols, gridList, rowList);
        for(i=0; i<grid.numCols; ++i){
            for(j=0; j<grid.numRows; ++j){
                tmpGrid.array[i][j] = factor*grid.array[j][grid.numCols - 1 - i];
            }
        }
        break;
    }

    return tmpGrid;
}

int intGridFindClosestTo(IntGrid g, int* x, int* y){
    int xdir = 0;
    int ydir = 1;
    int tmp;
    int counter = 0;
    int nextCorner = 0;
    int lengthChanger = 1;
    int length = 0;
    while(g.array[*x][*y] < 0){
        if(counter == nextCorner){
            tmp = ydir;
            ydir = xdir;
            xdir = -tmp;
            if(lengthChanger){
                length += 1;
                lengthChanger = 0;
            }
            else
                lengthChanger = 1;
            nextCorner += length;
        }
        counter++;
        *x += xdir;
        *y += ydir;
        if(*x >= g.numCols || *x < 0 || *y < 0 || *y >= g.numRows)
            return 0;
    }
    return 1;
}

void intGridFindBest8x8(IntGrid *grid, int* col, int* row){
    int maxIdx;
    IntGrid localSumGrid = intGridLocalSum(grid);
    intGridMaximum(&localSumGrid, &maxIdx);
    intGridIndToSub(&localSumGrid, maxIdx, col, row);
}

IntGrid intGridLocalSum(IntGrid *g) {
    unsigned char windowSize = DECODING_GRID_LENGTH;
    unsigned short windowSizeMinusOne = windowSize - 1;
    unsigned char i, j;
    unsigned int accumulator;
    int* dataPtrIn, *dataPtrOut;
    int gridList[MAX_NUM_POINTS];
    int* rowList[MAX_GRID_LENGTH];
    IntGrid tmp, summedGrid;
    tmp = intGridCreate(g->numCols - windowSize + 1, g->numRows, gridList, rowList);
    for(i=g->numRows; i--;){
        dataPtrIn = g->array[i];
        dataPtrOut = tmp.array[i];
        accumulator = 0;
        accumulator = dataPtrIn[0] +
                dataPtrIn[1] +
                dataPtrIn[2] +
                dataPtrIn[3] +
                dataPtrIn[4] +
                dataPtrIn[5] +
                dataPtrIn[6];

        for(j=0; j<tmp.numCols; j++){
            accumulator += dataPtrIn[j + windowSizeMinusOne];
            *(dataPtrOut++) = accumulator;
            accumulator -= dataPtrIn[j];
        }
    }

    //Recycle max probability grid data since we have no more need for it.
    summedGrid = intGridCreate(tmp.numCols, tmp.numRows - windowSize + 1, g->data, g->array);
    for(i=tmp.numCols; i--;){
        accumulator = tmp.array[0][i] +
                tmp.array[1][i] +
                tmp.array[2][i] +
                tmp.array[3][i] +
                tmp.array[4][i] +
                tmp.array[5][i] +
                tmp.array[6][i];

        for(j=0; j<summedGrid.numRows; j++){
            accumulator += tmp.array[j + windowSizeMinusOne][i];
            summedGrid.array[j][i] = accumulator;
            accumulator -= tmp.array[j][i];
        }
    }
    return summedGrid;
}

#pragma GCC push_options

#ifdef __cplusplus
}
#endif
